// =============================================================================
// Native Funktionen zu de.auerswald.lib.transfer.usb.AuerUSBChannel.
// 
// Diese Version verwendet libusb und ist damit ziemlich portabel.
// =============================================================================

// Autor:   Wolfgang Muees
// Stand:   15.3.2012
// Version: 2.0.8

// -----------------------------------------------------------------------------
// Fuer die Dauer der Entwicklung...
#undef DEBUG

// -----------------------------------------------------------------------------
// Includes
#include "de_auerswald_lib_transfer_usb_AuerUSBChannel.h"	// JNI Header

#include <stdio.h>	// standard IO functions
#include <stdlib.h>	// standard C functions
#include <usb.h>	// libusb header
#include <string.h>	// memcpy and friends
#include <errno.h>	// error codes

// -----------------------------------------------------------------------------
// Debug-Unterstuetzung
#ifdef DEBUG
#define dbg(format, arg...) printf(format "\n" , ## arg)
#else
#define dbg(format, arg...) do {} while (0)
#endif

// -----------------------------------------------------------------------------
// Properties
#define AUPMAXTRANSFERSIZE 0x01		// gibt die maximale TransferSize
#define AUPTRANSFEROFFSET  0x02		// gibt den TransferOffset zurueck

//-----------------------------------------------------------
#define MAXDEVICES 	16
#define MAXNAMELEN	100	 // Zeichen sollten ausreichen, oder?
#define MAXDEVLEN	20

#define ID_AUERSWALD	0x09BF	 // USB vendor ID of Auerswald

/*-------------------------------------------------------------------*/
/* vendor specific protocol                                          */
/* Header Byte */
#define AUH_INDIRMASK   0x80    /* mask for direct/indirect bit */
#define AUH_DIRECT      0x00    /* data is for USB device */
#define AUH_INDIRECT    0x80    /* USB device is relay */

#define AUH_SPLITMASK   0x40    /* mask for split bit */
#define AUH_UNSPLIT     0x00    /* data block is full-size */
#define AUH_SPLIT       0x40    /* data block is part of a larger one,
                                   split-byte follows */
#define AUH_SYNC	0x40	/* Sync to start of HDLC frame for B1,B2 */

#define AUH_TYPEMASK    0x3F    /* mask for type of data transfer */
#define AUH_TYPESIZE    0x40    /* different types */
#define AUH_DCHANNEL    0x00    /* D channel data */
#define AUH_B1CHANNEL   0x01    /* B1 channel transparent */
#define AUH_B2CHANNEL   0x02    /* B2 channel transparent */
/*                0x03..0x0F       reserved for driver internal use */
#define AUH_COMMAND     0x10    /* Command channel */
#define AUH_BPROT       0x11    /* Configuration block protocol */
#define AUH_DPROTANA    0x12    /* D channel protocol analyzer */
#define AUH_TAPI        0x13    /* telephone api data (ATD) */
/*                0x14..0x3F       reserved for other protocols */
#define AUH_UNASSIGNED  0xFF    /* if char device has no assigned service */
#define AUH_FIRSTUSERCH 0x11    /* first channel which is available for driver users */

#define AUH_SIZE	1 	/* Size of Header Byte */

/* Split Byte. Only present if split bit in header byte set.*/
#define AUS_STARTMASK   0x80    /* mask for first block of splitted frame */
#define AUS_FIRST       0x80    /* first block */
#define AUS_FOLLOW      0x00    /* following block */

#define AUS_ENDMASK     0x40    /* mask for last block of splitted frame */
#define AUS_END         0x40    /* last block */
#define AUS_NOEND       0x00    /* not the last block */

#define AUS_LENMASK     0x3F    /* mask for block length information */

/* Request types */
#define AUT_RREQ        (USB_ENDPOINT_IN  | USB_TYPE_VENDOR | USB_RECIP_OTHER)   /* Read Request */
#define AUT_WREQ        (USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_OTHER)   /* Write Request */

/* Vendor Requests */
#define AUV_GETINFO     0x00    /* GetDeviceInfo */
#define AUV_WBLOCK      0x01    /* Write Block */
#define AUV_RBLOCK      0x02    /* Read Block */
#define AUV_CHANNELCTL  0x03    /* Channel Control */
#define AUV_DUMMY	0x04	/* Dummy Out for retry */

/* Device Info Types */
#define AUDI_NUMBCH     0x0000  /* Number of supported B channels */
#define AUDI_OUTFSIZE   0x0001  /* Size of OUT B channel fifos */
#define AUDI_MBCTRANS   0x0002  /* max. Blocklength of control transfer */

/* Interrupt endpoint definitions */
#define AU_IRQENDP      1       /* Endpoint number */
#define AU_IRQCMDID     16      /* Command-block ID */
#define AU_BLOCKRDY     0       /* Command: Block data ready on ctl endpoint */
#define AU_IRQMINSIZE	5	/* Nr. of bytes decoded in this driver */
#define AU_IRQINTERVAL  1	/* Poll as fast as possible */	

/* Device String Descriptors */
#define AUSI_VENDOR   	1	/* "Auerswald GmbH & Co. KG" */
#define AUSI_DEVICE   	2	/* Name of the Device */
#define AUSI_SERIALNR 	3	/* Serial Number */
#define AUSI_MSN      	4	/* "MSN ..." (first) Multiple Subscriber Number */

#define AUSI_DLEN	100	/* Max. Length of Device Description */

#define AU_WRITE_TIMEOUT 100    // ms to transfer a write control paket (can be long, nothing serious)
#define	AU_READ_TIMEOUT	 250    // ms to transfer a read control paket (can be long, nothing serious)
#define AU_POLL_TIMEOUT  1350   // ms to check if a read control paket is on the way (polling)

#define AU_CONFIG	1	// Configuration
#define AU_INTERFACE	0	// Interface

#define WRITE_RETRIES	3	// do some retries
#define READ_RETRIES	1	// wait only once

// Hold internal information about the USB device and communication state
struct auer_dev {
	usb_dev_handle *dev_handle;	// handle for open device of usblib
	int maxControlLength;		// maximum length of a control transfer
	int channel;			// number of open channel
	int version;			// Version code of the device
	int readcount;			// Number of bytes in last read
	jbyte *buffer;			// Databuffer for read
};

// Array of pointers to the device structs
// static initialized to NULL
static struct auer_dev *dev_array[MAXDEVICES];

// -----------------------------------------------------------------------------
// Field IDs in the JAVA object
static jfieldID RawFieldID;		// ID to store the address of auer_dev
static jfieldID MaxTransferFieldID;	// ID to store the max. Transfer Length

// -----------------------------------------------------------------------------
// Store the address of the device struct in a global array, and store the
// array index in the JAVA channel object.
static inline int setup_auerdev(JNIEnv *env, jobject obj, struct auer_dev *auerdev)
{
    jint i;
    for (i = 0; i < MAXDEVICES; i++) {
        if (!dev_array[i]) {
            dev_array[i] = auerdev;
        	RawFieldID = env->GetFieldID( env->GetObjectClass( obj), "CUsbRawTransmitPtr", "I");
	        env->SetIntField( obj, RawFieldID, i+1);    // to avoid the "0"
            return 0;
        }
    }
    dbg("can not store device struct in global array");
    return -ENOMEM;
}

// Clear the address of the device struct in the global array, and clear the
// array index in the JAVA channel object.
static inline void clear_auerdev(JNIEnv *env, jobject obj)
{
    jint i = env->GetIntField( obj, RawFieldID);
    if (i > 0) {
        env->SetIntField( obj, RawFieldID, 0);
        dev_array[i-1] = NULL;
    }
}

// Retrieve the array index from the JAVA channel object, and get the
// address of the device struct out of the global array.
static inline struct auer_dev * retrieve_auerdev(JNIEnv *env, jobject obj)
{
    jint i = env->GetIntField( obj, RawFieldID);
    if (i > 0) 
        return dev_array[i-1];
    else
        return NULL;
}

// -----------------------------------------------------------------------------
// Performs a control request.
// Returns number of bytes written/read or < 0 on error.
static int auer_control_msg(struct auer_dev *auerdev, int requesttype, int request, int value,
			    int index, char *bytes, int size, int timeout, int retries)
{
	int retval = 0;
	
	if (!auerdev) {
		dbg("auerdev == NULL in auer_control_msg");
		return -ENXIO;
	}

	for (; retries; retries--) {
		retval = usb_control_msg(auerdev->dev_handle, requesttype, request, value, index, bytes, size, timeout);
		if (retval >= 0)
			return retval;
	}
	return retval;
}

// -----------------------------------------------------------------------------
// Unregister a channel
static void auer_release_channel(struct auer_dev *auerdev, int channel)
{
	int ret;

	// valid channel?
	if (!auerdev)
		return;
	// Channel ueberpruefen
	if ((channel < AUH_FIRSTUSERCH) || (channel >= AUH_TYPESIZE))
		return;

	ret = auer_control_msg(
		auerdev, 			        /* pointer to device */
		AUT_WREQ,                               /* USB message request type value */
		AUV_CHANNELCTL,                         /* USB message request value */
		0x00,              // close             /* USB message value */
		auerdev->channel,		        /* USB message index value */
		NULL,                                   /* pointer to the data to send */
		0,                                      /* length in bytes of the data to send */
		AU_WRITE_TIMEOUT,			/* Timeout in ms */
		WRITE_RETRIES);
	auerdev->channel = AUH_UNASSIGNED;
	ret = ret;
	dbg("Release Channel return code = %d", ret);
}

// Register a channel
static int auer_register_channel(struct auer_dev *auerdev, int channel)
{
	int ret;

	// Channel ueberpruefen
	if ((channel < AUH_FIRSTUSERCH) || (channel >= AUH_TYPESIZE)) {
		dbg ("Invalid ChannelTyp %d", channel);
		return -EINVAL;
	}

	/* register service in device */
	ret = auer_control_msg(
		auerdev,                                /* pointer to device */
		AUT_WREQ,                               /* USB message request type value */
		AUV_CHANNELCTL,                         /* USB message request value */
		0x01,              /* open                 USB message value */
		channel,            		        /* USB message index value */
		NULL,                                   /* pointer to the data to send */
		0,                                      /* length in bytes of the data to send */
		AU_WRITE_TIMEOUT,			/* Timeout in ms */
		WRITE_RETRIES);
	if (ret < 0) {
		dbg ("Register Channel failed with error code %d", ret);
		return -EIO;
	}
	dbg("Channel %d registered", channel);
	auerdev->channel = channel;
	return 0;
}


// Do a read. Number of bytes go to readcount, data in buffer
static int auer_read(struct auer_dev *auerdev) 
{
	int ret;

	// Daten einlesen
	ret = auer_control_msg(
		auerdev, 			        /* pointer to device */
		AUT_RREQ,                               /* USB message request type value */
		AUV_RBLOCK,                             /* USB message request value */
		0x00,                                   /* USB message value */
		auerdev->channel | AUH_DIRECT | AUH_UNSPLIT, /* USB message index value */
		(char *)auerdev->buffer,                /* pointer to the data to receive */
		auerdev->maxControlLength+AUH_SIZE,     /* length in bytes of the data to receive */
		AU_POLL_TIMEOUT,	                /* Timeout in ms */
		READ_RETRIES);
	if (ret <= 0) {
		dbg("Konnte keine Daten einlesen, Status = %d", ret);
		auerdev->readcount = 0;
		return -EIO;
	}
	if ((*auerdev->buffer & AUH_TYPEMASK) != auerdev->channel) {
		dbg("Data for wrong channel received");
		auerdev->readcount = 0;
		return -EIO;
	}
	// we have data
	auerdev->readcount = ret;
	return ret;
}


// #############################################################################
// # ========================== INTERFACE-FUNCTIONS: ========================= #
// #              Die Beschreibung der Funktionen stehen im Java Teil.	       #
// #############################################################################


/*
 * Class:     de_auerswald_lib_transfer_usb_AuerUSBChannel
 * Method:    getDeviceList
 * Signature: ()[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL Java_de_auerswald_lib_transfer_usb_AuerUSBChannel_getDeviceList
  (JNIEnv *env, jclass classObj)
{
	jobjectArray temp = NULL;
	jstring s = NULL;
	jclass stringClass = NULL;
	struct usb_bus *bus;
	unsigned int u;
	unsigned int count;
	char devnames[MAXDEVICES][MAXDEVLEN];
	char ufnames [MAXDEVICES][MAXNAMELEN];
	int ret;

	// Syncronized Teil 1
	env->MonitorEnter(classObj);

        usb_init();
        usb_find_busses();
        usb_find_devices();
        bus = usb_get_busses();

	count = 0;
        for (; bus; bus = bus->next) {
                struct usb_device *dev;
                for (dev = bus->devices; dev; dev = dev->next) {
			if (dev->descriptor.idVendor == ID_AUERSWALD) {
				usb_dev_handle *dev_handle;
				u = 0;

				switch (dev->descriptor.idProduct) {
				case 0x00C0:          /* COMpact 2104 USB/DSL */
				case 0x00DB:          /* COMpact 4410/2206 USB */
				case 0x00DC:          /* COMpact 4406 DSL */
				case 0x00DD:          /* COMpact 2204 USB */
				case 0x00DE:          /* COMpact 2104.2 USB */
				case 0x00F1:          /* COMfort 2000 System Telephone */
				case 0x00F2:          /* COMfort 1200 System Telephone */
				case 0x00F5:          /* COMfortel 2500 System Telephone */
				case 0x00FB:	      /* COMfortel 2600 System Telephone */
				// add more devices HERE
					dev_handle = usb_open(dev);
					if (dev_handle) {
						sprintf( devnames[count], "%p", dev);
						/* Try to get a suitable textual description of the device */
						/* Device name:*/
						ret = usb_get_string_simple(dev_handle, AUSI_DEVICE, ufnames[count], MAXNAMELEN-1);
						if (ret > 0) {
							u += ret;
							/* Append Serial Number */
							memcpy(&ufnames[count][u], ",Ser# ", 6);
							u += 6;
							ret = usb_get_string_simple(dev_handle, AUSI_SERIALNR, &ufnames[count][u], MAXNAMELEN-u-1);
							if (ret > 0) {
								u += ret;
								/* Append subscriber number */
								memcpy(&ufnames[count][u], ", ", 2);
								u += 2;
								ret = usb_get_string_simple( dev_handle, AUSI_MSN, &ufnames[count][u], MAXNAMELEN-u-1);
								if (ret > 0) {
									u += ret;
								}
							}
						}
						usb_close(dev_handle);
					}
					ufnames[count][u] = '\0';
					dbg ("Device %s ist ein %s", devnames[count], ufnames[count]);
					count++;
					if (count >= MAXDEVICES)
						goto endloop;
					break;
				default:
					break;
				}
			}
		}
	}
endloop:

	// Haben wir Namen gefunden?
	if (count == 0) {
		dbg ("Keine Geraete am Bus gefunden");
		goto fail;
	}

	// Die String Class besorgen
	stringClass = env->FindClass ("java/lang/String");
	if (stringClass == NULL) {
		goto FINAL;
	}
	// Jetzt StringArray Allocieren
	temp =  env->NewObjectArray(count*2,stringClass, NULL);

	if(NULL == temp) {
		goto FINAL;
	}

	// So, jetzt haben wir das Array
	// Array mit String's fllen
	for(u = 0 ; u < count ; u++) {
		// Driver Namen schreiben
		s =  env->NewStringUTF( &devnames[u][0]);
		if(NULL == s)
		{
			temp = NULL;
			goto FINAL;
		}
		// das neue String Objekt ins Array copieren
		env->SetObjectArrayElement (temp, u*2, s);

		// Friendly Namen schreiben
		s =  env->NewStringUTF( &ufnames[u][0]);
		if(NULL == s)
		{
			temp = NULL;
			goto FINAL;
		}
		// das neue String Objekt ins Array copieren
		env->SetObjectArrayElement(temp, u*2+1, s);
	}
	goto FINAL;
// ..........................
fail:	{
		// alte Exception loeschen
		env->ExceptionClear();

		jclass newExcCls = env->FindClass("de/auerswald/lib/transfer/usb/UsbNoDeviceException");
		if (newExcCls == 0) {
			temp = NULL;
			goto FINAL;
		}
		env->ThrowNew(newExcCls, "thrown from C code");
	}
// ..........................
FINAL:	// Syncronized Teil 2 und Ende
	env->MonitorExit(classObj);
	// Jetzt die Namen der Devices als String Array zurckgeben
	return temp;
}


/*
 * Class:     de_auerswald_lib_transfer_usb_AuerUSBChannel
 * Method:    openChannel
 * Signature: (Ljava/lang/String;I)I
 */
JNIEXPORT void JNICALL Java_de_auerswald_lib_transfer_usb_AuerUSBChannel_openChannel
  (JNIEnv *env, jobject obj, jstring DeviceName, jint ChannelTyp, jint RingBufferSize)
{
	const char *pDeviceName = NULL;
        struct usb_device *dev = NULL;
	struct auer_dev *auerdev = NULL;
	unsigned char *buf = NULL;
	int ret;

	dbg ("openChannel called" );

	// ID des Java-Speichers ermitteln
	MaxTransferFieldID = env->GetFieldID( env->GetObjectClass( obj), "maxUseableTransferSize", "I");

	// Devicenamen berprfen
	pDeviceName = env->GetStringUTFChars(DeviceName, 0);
	if (!pDeviceName) {
		dbg ("Devicename nicht gefunden");
		goto fail;
	}
	sscanf(pDeviceName, "%p", &dev);
	if (!dev) {
		dbg ("Devicename == NULL");
		goto fail;
	}

	// Allocate Memory for auer_dev
	auerdev = (struct auer_dev *) malloc(sizeof(struct auer_dev));
	if (!auerdev) {
		dbg ("malloc (auerdev) failed");
		goto fail;
	}
	auerdev->dev_handle = NULL;
	auerdev->maxControlLength = 0;
	auerdev->channel = AUH_UNASSIGNED;
	auerdev->version = 0;
	auerdev->readcount = 0;
	auerdev->buffer = NULL;

	// �fne das Device
	auerdev->dev_handle = usb_open(dev);
	if (!auerdev->dev_handle) {
		dbg ("Device konnte nicht geoeffnet werden");
		goto fail;
	}
	dbg("usb_open() erfolgreich");

	// remove the kernel driver
#if LIBUSB_HAS_DETACH_KERNEL_DRIVER_NP
	(void) usb_detach_kernel_driver_np(auerdev->dev_handle, 0);
#endif

	/* Get the usb version of the device */
	auerdev->version = dev->descriptor.bcdDevice;
	dbg ("Version is %X", auerdev->version);

	// Set the configuration
        (void) usb_set_configuration(auerdev->dev_handle, AU_CONFIG);

	// Set the interface
	(void) usb_claim_interface(auerdev->dev_handle, AU_INTERFACE);

        /* requesting a service means: release the previous one first */
	auer_release_channel(auerdev, ChannelTyp);

	/* register service in device */
	ret = auer_register_channel(auerdev, ChannelTyp);
	if (ret < 0) {
		goto fail;
	}

        /* get the maximum allowed control transfer length */
	buf = (unsigned char *) malloc(2);
	if (!buf) {
		dbg ("Buffer allocation failed");
		goto fail;
	}
	ret = auer_control_msg(
		auerdev,                            /* pointer to device */
                AUT_RREQ,                           /* USB message request type value */
                AUV_GETINFO,                        /* USB message request value */
                0,                                  /* USB message value */
                AUDI_MBCTRANS,                      /* USB message index value */
                (char *)buf,                        /* pointer to the receive buffer */
                2,                                  /* length of the buffer */
		AU_READ_TIMEOUT,                    /* Timeout in ms */
		READ_RETRIES);
	if (ret == 2) {
		auerdev->maxControlLength = buf[1];
		auerdev->maxControlLength <<= 8;
		auerdev->maxControlLength |= buf[0];
                dbg("max. allowed control transfersize is %d bytes", auerdev->maxControlLength);
        } else {
		dbg("getting max. allowed control transfer length failed with error %d", ret);
		goto fail;
        }

	// allocate the read buffer
	auerdev->buffer = (jbyte *) malloc(auerdev->maxControlLength+AUH_SIZE);
	if (!auerdev->buffer) {
		dbg("no memory for read buffer");
		goto fail;
	}

	// store address of auerdev in the JAVA object
	if (setup_auerdev(env, obj, auerdev) < 0) 
        goto fail;

	// setup the max. Transfer Length in the JAVA object
	env->SetIntField( obj, MaxTransferFieldID, (jint) auerdev->maxControlLength);

	goto FINAL;
// ...............................
fail:	// Kanal zum Treiber wieder aufheben
	if (auerdev) {
		if (auerdev->buffer)
			free(auerdev->buffer);
		if (auerdev->dev_handle) {
			auer_release_channel(auerdev, auerdev->channel);
			(void) usb_release_interface(auerdev->dev_handle, AU_INTERFACE);
			usb_close(auerdev->dev_handle);
		}
		free(auerdev);
	} 

	// alte Exception loeschen
	env->ExceptionClear();
	{
		jclass newExcCls = env->FindClass("de/auerswald/lib/transfer/usb/UsbInitDeviceException");
		if (newExcCls == 0)
		{ // Unable to find the new exception class, give up.
			goto FINAL;
		}
		env->ThrowNew(newExcCls, "thrown from C code");
	}
// ...............................
FINAL:	// Java-Resourcen wieder freigeben
	if (pDeviceName)
		env->ReleaseStringUTFChars(DeviceName, pDeviceName);
	// Buffer wird nicht mehr gebraucht
	if (buf)
		free(buf);
}


/*
 * Class:     de_auerswald_lib_transfer_usb_AuerUSBChannel
 * Method:    closeChannel
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_de_auerswald_lib_transfer_usb_AuerUSBChannel_closeChannel
  (JNIEnv * env, jobject obj)
{
	struct auer_dev *auerdev = retrieve_auerdev(env, obj);
	// schliesse den Kanal
	if (auerdev) {
		if (auerdev->buffer)
			free(auerdev->buffer);
		if (auerdev->dev_handle) {
		        // release requested service
			auer_release_channel(auerdev, auerdev->channel);
			(void) usb_release_interface(auerdev->dev_handle, AU_INTERFACE);
			usb_close(auerdev->dev_handle);
		}
		free(auerdev);
	}
	// und vermerke dies im Java-Objekt
	clear_auerdev(env, obj);
}


/*
 * Class:     de_auerswald_lib_transfer_usb_AuerUSBChannel
 * Method:    getProperty
 * Signature: (I)[I
 */
JNIEXPORT jint JNICALL Java_de_auerswald_lib_transfer_usb_AuerUSBChannel_getProperty
  (JNIEnv *env, jobject obj, jint PropertyTyp,jintArray Data)
{
	struct auer_dev *auerdev = retrieve_auerdev(env, obj);
	unsigned int u = 0;
	jint DataLen = 0;

	if(!auerdev) {
		// Exception kein Channel da
		env->ExceptionClear();
		jclass newExcCls = env->FindClass("de/auerswald/lib/transfer/usb/UsbNoLineException");
		if (newExcCls == 0)
		{ // Unable to find the new exception class, give up.
			return 0;
		}
		env->ThrowNew(newExcCls, "thrown from C code");
		return 0;
	}

	// --- JNI-IntArray erstmal konvertieren und nutzbar machen: ---
	jint *RealBuffer = env->GetIntArrayElements(Data, 0);

	// Was fr eine Property ist es denn?
	switch ( PropertyTyp) {
	case AUPMAXTRANSFERSIZE: 	// gibt die maximale TransferSize
		u = auerdev->maxControlLength;
		dbg("max. Buffersize = %d", u);
		break;
	case AUPTRANSFEROFFSET:		// gibt den TransferOffset zurck
		u = 0;			// sowas krudes haben wir nicht
		break;
	default:			// unbekannte property
		goto fail;
	}
	// Ergebnis in JAVA-Variable ablegen
	*RealBuffer = u;
	DataLen = 1;

	goto FINAL;
// ..................
fail:	// alte Exception loeschen
	env->ExceptionClear();
	{
		jclass newExcCls = env->FindClass("de/auerswald/lib/transfer/usb/UsbInvalidPropertyException");
		if (newExcCls == 0)
		{ // Unable to find the new exception class, give up.
			goto FINAL;
		}
		env->ThrowNew(newExcCls, "thrown from C code");
	}
// ..................
FINAL:	// Java-Variable wieder freigeben
	env->ReleaseIntArrayElements(Data, RealBuffer, 0);

	// Anzahl der gesetzten Werte zurueckgeben
	return DataLen;
}


/*
 * Class:     de_auerswald_lib_transfer_usb_AuerUSBChannel
 * Method:    setProperty
 * Signature: (I[I)V
 */
JNIEXPORT void JNICALL Java_de_auerswald_lib_transfer_usb_AuerUSBChannel_setProperty
  (JNIEnv *env, jobject obj, jint PropertyTyp, jintArray value)
{
	struct auer_dev *auerdev = retrieve_auerdev(env, obj);

	if(!auerdev) {
		// Exception kein Channel da
		env->ExceptionClear();
		jclass newExcCls = env->FindClass("de/auerswald/lib/transfer/usb/UsbNoLineException");
		if (newExcCls == 0)
		{ // Unable to find the new exception class, give up.
			return;
		}
		env->ThrowNew(newExcCls, "thrown from C code");
		return;
	}
	// Leider gibt es z.Z. gar keine Properties zu setzen ...
	goto fail;
// ....................
fail:	// alte Exception loeschen
	env->ExceptionClear();
	{
		jclass newExcCls = env->FindClass("de/auerswald/lib/transfer/usb/UsbInvalidPropertyException");
		if (newExcCls == 0)
		{ // Unable to find the new exception class, give up.
			goto FINAL;
		}
		env->ThrowNew(newExcCls, "thrown from C code");
	}
// ..................
FINAL:	// nix
	return;
}


/*
 * Class:     de_auerswald_lib_transfer_usb_AuerUSBChannel
 * Method:    write
 * Signature: ([BIIJ)V
 */
JNIEXPORT void JNICALL Java_de_auerswald_lib_transfer_usb_AuerUSBChannel_write
  (JNIEnv * env , jobject obj, jbyteArray data, jint offset, jint length, jint flags)
{
	struct auer_dev *auerdev = retrieve_auerdev(env, obj);
	int ret = 0;
	int len = length;
	char *buffer = NULL;
	jbyte *RealBuffer = NULL;

	// check for device
	if(!auerdev) {
		dbg ("no device in write()");
		goto fail;
	}

	// check for length
	if (length > auerdev->maxControlLength) {
		dbg ("length too long: %d Bytes, Max = %d - truncated", (int)length, auerdev->maxControlLength);
		len = auerdev->maxControlLength;
	}

	// allocate data buffer
	buffer = (char *) malloc(len+AUH_SIZE);
	if (!buffer) {
		dbg("no memory write()");
		goto fail;
	}

	// --- JNI-ByteArray erstmal konvertieren und nutzbar machen: ---
	RealBuffer = env->GetByteArrayElements(data, NULL);
	memcpy(buffer+AUH_SIZE, &RealBuffer[offset], len);
	env->ReleaseByteArrayElements(data, RealBuffer, 0);

	/* set the header byte */
        *buffer = auerdev->channel | AUH_DIRECT | AUH_UNSPLIT;

	dbg("Schreibe %d Bytes", len);

	/* send URB synchronous */
	ret = auer_control_msg(
		auerdev,  			    /* pointer to device */
                AUT_WREQ,                           /* USB message request type value */
                AUV_WBLOCK,                         /* USB message request value */
                0,                                  /* USB message value */
                auerdev->channel | AUH_DIRECT | AUH_UNSPLIT, /* USB message index value */
                buffer,                             /* pointer to the send buffer */
                len+AUH_SIZE,                       /* length of the buffer */
		AU_WRITE_TIMEOUT,                   /* Timeout in ms */
		WRITE_RETRIES);
	if (ret < len+AUH_SIZE) {
		// nicht alle Daten konnten geschrieben werden
		dbg ("write gab %d zurueck", ret);
		goto fail;
	}

	goto FINAL;
// ...........................
fail:	// alte Exception loeschen
	env->ExceptionClear();
	{
		jclass newExcCls = env->FindClass("de/auerswald/lib/transfer/usb/UsbNoLineException");
		if (newExcCls == 0)
		{ // Unable to find the new exception class, give up.
			goto FINAL;
		}
		env->ThrowNew(newExcCls, "thrown from C code");
	}
// ...........................
FINAL:	if (buffer)
		free(buffer);

	return;
}


/*
 * Class:     de_auerswald_lib_transfer_usb_AuerUSBChannel
 * Method:    readDataAvailable
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_de_auerswald_lib_transfer_usb_AuerUSBChannel_readDataAvailable
  (JNIEnv * env , jobject obj)
{
	struct auer_dev *auerdev = retrieve_auerdev(env, obj);
	int ret = 0;

	if(!auerdev) {
		// Exception kein Channel da
		env->ExceptionClear();
		jclass newExcCls = env->FindClass("de/auerswald/lib/transfer/usb/UsbNoLineException");
		if (newExcCls == 0)
		{ // Unable to find the new exception class, give up.
			return false;
		}
		env->ThrowNew(newExcCls, "thrown from C code");
		return false;
	}

	// result of last read?
	if (auerdev->readcount > 0)
		return true;

	// polle das Device
	ret = auer_read(auerdev);
	return (ret > 0);
}


/*
 * Class:     de_auerswald_lib_transfer_usb_AuerUSBChannel
 * Method:    readNextBlock
 * Signature: ()[B
 */
JNIEXPORT jbyteArray JNICALL Java_de_auerswald_lib_transfer_usb_AuerUSBChannel_readNextBlock
  (JNIEnv * env , jobject obj )
{
	struct auer_dev *auerdev = retrieve_auerdev(env, obj);
	jbyteArray result = NULL;
	int ret = 0;
	
	// check for device
	if(!auerdev) {
		// Exception kein Channel da
		env->ExceptionClear();
		jclass newExcCls = env->FindClass("de/auerswald/lib/transfer/usb/UsbNoLineException");
		if (newExcCls == 0)
		{ // Unable to find the new exception class, give up.
			return NULL;
		}
		env->ThrowNew(newExcCls, "thrown from C code");
		return NULL;
	}

	ret = auerdev->readcount;
	if (ret <= 0) {
		dbg( "Konnte keine Daten lesen");
		goto fail;
	}

	dbg ("%d Bytes gelesen", ret);
	ret -= AUH_SIZE;

	// Java-Array erzeugen
	result = env->NewByteArray(ret);
	if(!result) {
		goto fail;
	}

	// Daten nach Java kopieren
	env->SetByteArrayRegion(result,0,ret,auerdev->buffer+AUH_SIZE);
	auerdev->readcount = 0;

	// Fertig
	return result;
// ..................
fail:	// Aufraeumen und Fehler melden
	// alte Exceptions l�chen
	env->ExceptionClear();
	jclass newExcCls = env->FindClass("de/auerswald/lib/transfer/usb/UsbNoDataException");
	if (newExcCls == 0)
	{ // Unable to find the new exception class, give up.
		return NULL;
	}
	env->ThrowNew(newExcCls, "thrown from C code");
	return NULL;
}


/*
 * Class:     de_auerswald_lib_transfer_usb_AuerUSBChannel
 * Method:    readNextBlock
 * Signature: ([B)V
 */
JNIEXPORT void JNICALL Java_de_auerswald_lib_transfer_usb_AuerUSBChannel_readNextBlock___3B
  (JNIEnv * env, jobject obj, jbyteArray result)
{
	struct auer_dev *auerdev = retrieve_auerdev(env, obj);
	int ret = 0;
	jint length = env->GetArrayLength(result);

	if(!auerdev) {
		// Exception kein Channel da
		env->ExceptionClear();
		jclass newExcCls = env->FindClass("de/auerswald/lib/transfer/usb/UsbNoLineException");
		if (newExcCls == 0)
		{ // Unable to find the new exception class, give up.
			return;
		}
		env->ThrowNew(newExcCls, "thrown from C code");
		return;
	}

	ret = auerdev->readcount;
	if (ret <= 0) {
		dbg( "Konnte keine Daten lesen");
		goto fail;
	}

	dbg ("%d Bytes gelesen", ret);
	ret -= AUH_SIZE;

	// Daten nach Java kopieren
	if (ret > length) {
		dbg ("%d Bytes read, %d Bytes free in Buffer", ret, (int)length);
		ret = length;
	}
	env->SetByteArrayRegion(result,0,ret,auerdev->buffer+AUH_SIZE);
	auerdev->readcount = 0;

	// Fertig
	return;
// ..................
fail:	// Aufraeumen und Fehler melden
	{
		// alte Exceptions loeschen
		env->ExceptionClear();
		jclass newExcCls = env->FindClass("de/auerswald/lib/transfer/usb/UsbNoDataException");
		if (newExcCls == 0)
		{ // Unable to find the new exception class, give up.
			return;
		}
		env->ThrowNew(newExcCls, "thrown from C code");
	}
}


/*
 * Class:     de_auerswald_lib_transfer_usb_AuerUSBChannel
 * Method:    isConnected
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_de_auerswald_lib_transfer_usb_AuerUSBChannel_isConnected
  (JNIEnv *env, jobject obj)
{
	struct auer_dev *auerdev = retrieve_auerdev(env, obj);

	// wenn der Kanal offen ist, sind wir verbunden
	if (auerdev)
		if (auerdev->channel != AUH_UNASSIGNED)
			return true;

	// sonst nicht
	return false;
}


/*
 * Class:     de_auerswald_lib_transfer_usb_AuerUSBChannel
 * Method:    isSuspended
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_de_auerswald_lib_transfer_usb_AuerUSBChannel_isSuspended
  (JNIEnv *, jobject)
{
	// das sind wir nie ... denn dann ist das Device auch nicht da
	return false;
}


/*
 * Class:     de_auerswald_lib_transfer_usb_AuerUSBChannel
 * Method:    TransmitterReady
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_de_auerswald_lib_transfer_usb_AuerUSBChannel_TransmitterReady
  (JNIEnv *env , jobject obj)
{
	struct auer_dev *auerdev = retrieve_auerdev(env, obj);

	// wenn der Kanal offen ist, koennen wir auch senden
	if (auerdev)
		if (auerdev->channel != AUH_UNASSIGNED)
			return true;

	// sonst nicht
	return false;
}

